/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "kite_object.h"
#include "vm/kite_vm.h"
#include "config.h"

/*****************************************************************************
 * Get environment variable
 ****************************************************************************/
KITE_CLASS_METHOD(Os_envGet) {
    (void)this;
    
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *name = kite_string_object(thd, arg1);

    char *res = getenv(name->builtin_data.stringvalue.string);
    if (res == NULL) {
        res = "";
    }
    kite_vm_return(thd, kite_new_string(thd, res));
}

#ifndef HAVE_SETENV
void setenv(char *name, char *value, int overwrite)
{
    (void)overwrite;
    char *ptr = malloc(strlen(name) + strlen(value) + 2);
    strcpy(ptr, name);
    strcat(ptr, "=");
    strcat(ptr, value);
    putenv(ptr);
    free(ptr);
}
#endif

/*****************************************************************************
 * Set environment variable
 ****************************************************************************/
KITE_CLASS_METHOD(Os_envSet) {
    (void)this;
    
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *arg2 = 
        args->builtin_data.listvalue.cdr->builtin_data.listvalue.car;
    kite_object_t *name = kite_string_object(thd, arg1);
    kite_object_t *value = kite_string_object(thd, arg2);

    setenv(name->builtin_data.stringvalue.string,
           value->builtin_data.stringvalue.string,
           1);
    kite_dereference_object(name);
    kite_dereference_object(value);
}

#ifndef WIN32
/*****************************************************************************
 * Fork a new process.
 ****************************************************************************/
KITE_CLASS_METHOD(Os_fork) {
    KITE_NO_ARGS;
    (void)this;
    int ret = fork();
    kite_vm_return(thd, kite_new_integer(thd, ret));
}
#endif

/*****************************************************************************
 * Get working directory
 ****************************************************************************/
KITE_CLASS_METHOD(Os_getcwd) {
    KITE_NO_ARGS; 
    (void)this;                                                                                     
                                                                                                       
    size_t cursize = 100;                                                                              
    char *path, *tmp;                                                                                  
    do {                                                                                               
        path = malloc(cursize);                                                                        
        tmp = getcwd(path, cursize);                                                                   
                                                                                                       
        if (tmp == path) break;
        else if (errno != ERANGE) {
            strcpy(path, "");
        }
        cursize *= 2;                                                                                  
    } while(errno == ERANGE);                                                                         
                                                                                                       
    kite_object_t *ret = kite_new_string(thd, path);                                                   
    free(path);                                                                                        
    kite_vm_return(thd, ret);
}

/*****************************************************************************
 * Change directory
 ****************************************************************************/
KITE_CLASS_METHOD(Os_chdir) {
    (void)this;
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *path = kite_string_object(thd, arg1);

    int ret = chdir(path->builtin_data.stringvalue.string);
    kite_dereference_object(path);
    kite_vm_return(thd, kite_new_boolean(thd, !ret));
}

/*****************************************************************************
 * Run arbitrary commands
 ****************************************************************************/
KITE_CLASS_METHOD(Os_system) {
    (void)this;
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *path = kite_string_object(thd, arg1);

    int ret = system(path->builtin_data.stringvalue.string);
    kite_dereference_object(path);
    kite_vm_return(thd, kite_new_integer(thd, ret));
}

KITE_MODULE_INITIALIZER(os) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *newclass = kite_new_class(thread, NULL, "os");
    kite_set_docstring(newclass, "Operating system functions.");

    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "envGet", 
        kite_new_method_with_docs(thread, FUNC_COMPILED, Os_envGet, 
                                  "Retrieve contents of environment variable.", 1,
                                  "var", "Environment variable to return."));
    kite_add_method(thread, newclass, "envSet", 
        kite_new_method_with_docs(thread, FUNC_COMPILED, Os_envSet, 
                                  "Set contents of environment variable to given value.", 2,
                                  "var", "Variable to set.",
                                  "val", "Value to set variable to."));
#ifndef WIN32
    kite_add_method(thread, newclass, "fork", 
        kite_new_method_with_docs(thread, FUNC_COMPILED, Os_fork, 
                                  "Forks off a new process.", 0));
#endif
    kite_add_method(thread, newclass, "chdir", 
        kite_new_method_with_docs(thread, FUNC_COMPILED, Os_chdir, 
                                  "Changes the current working directory.", 1,
                                  "newcwd", "The directory to change to."));
    kite_add_method(thread, newclass, "getcwd", 
        kite_new_method_with_docs(thread, FUNC_COMPILED, Os_getcwd, 
                                  "Gets the current working directory.", 0));
    kite_add_method(thread, newclass, "system", 
        kite_new_method_with_docs(thread, FUNC_COMPILED, Os_system, 
                                  "Executes system command.", 1,
                                  "cmd", "Command to execute."));

    /* Initialize args */
    char **tmp = thread->vm->args;
    kite_object_t *args = kite_new_list(thread);
    kite_object_t *tmpobj = args;
    while(tmp && *tmp) {
        kite_object_t *newarg = kite_new_string(thread, *tmp);
        if (tmpobj->builtin_data.listvalue.car) {
            tmpobj->builtin_data.listvalue.cdr = kite_new_list(thread);
            tmpobj = tmpobj->builtin_data.listvalue.cdr;
        }
        tmpobj->builtin_data.listvalue.car = newarg;
        tmp++;
    }
    
    kite_add_property(thread, newclass, "args", 1, "Arguments passed in from the shell.");
    kite_set_property(thread, newclass, "args", args);  
    kite_add_property(thread, newclass, "dylib_extension", 1, "The dynamic library extension on this system.");
    kite_set_property(thread, newclass, "dylib_extension", kite_new_string(thread, DYLIB_EXTENSION));
    
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "os", 1, NULL);
    kite_set_property(thread, parent, "os", newclass);
}
