#ifndef KITE_OBJS__KITEOBJECT_H
#define KITE_OBJS__KITEOBJECT_H
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#ifdef WIN32
#undef __MINGW32__
#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x501
#include <windows.h>
#endif
#include <pthread.h>

/*
 * FALSE/TRUE definitions (for those platforms that don't already define
 * them).
 */
#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

/* Dynamic library extensions. */
#ifdef WIN32
#define DYLIB_EXTENSION ".dll"
#else
#ifdef __APPLE__
#define DYLIB_EXTENSION ".dylib"
#else
#define DYLIB_EXTENSION ".so"
#endif
#endif

/*
 * Representation of Kite method written in C
 */
struct kite_object_t;
struct kite_thread_t;
typedef void (*kite_compiled_func_t)(struct kite_thread_t *, 
                                     struct kite_object_t *, 
                                     struct kite_object_t *);

/*
 * Representation of Kite method (built-in and user provided)
 */
typedef struct kite_function_t {
    enum { FUNC_COMPILED, 
           FUNC_BYTECODE } functype;
    void *funcptr;
    struct kite_object_t *this;
    int numargs;
    struct kite_object_t *arginfo;
} kite_function_t;

/*
 * Kite symbol table
 */
typedef struct kite_symtab_t
{
    char *name;
    char *docstring;
    int global;
    struct kite_symtab_t *left, *right;
    struct kite_object_t *value;
} kite_symtab_t;

#include "vm/kite_vm.h"
/*
 * Representation of an object/class in the Kite object system.
 */
typedef struct kite_object_t {
    unsigned int refcount;
    kite_list_t *gc_entry;
    enum { OBJ_NULL,
           OBJ_INTEGER, 
           OBJ_FLOAT,
           OBJ_STRING,
           OBJ_IDENT, 
           OBJ_LIST, 
           OBJ_BOOLEAN,
           OBJ_METHOD, 
           OBJ_CLASS, 
           OBJ_INSTANCE,
           } type;
    kite_thread_t *owner_thread;
    int shareable;
    struct kite_object_t *parent;
    union {
        long intvalue;
        double floatvalue;
        struct {
            int length;
            char *string;
        } stringvalue;
        struct {
            struct kite_object_t *car; /* current list element */
            struct kite_object_t *cdr; /* next list element */
            struct kite_object_t *iterator; /* current iterator position */
        } listvalue;
        kite_function_t funcvalue;
        kite_thread_t *threadvalue;
        FILE *filevalue;
    } builtin_data;
    struct {
        char *name;
        char *docstring;
        struct kite_object_t *inherit_from;
        struct kite_object_t *operators[NUM_OPERATORS];
        kite_symtab_t *properties;
        struct kite_object_t *destructor;
        int creatable;
        pthread_mutex_t mtx;
#ifdef WIN32
        HMODULE dylib_hndl;
#else
        void *dylib_hndl;
#endif
    } object_data;
} kite_object_t;

/*
 * General object creation/destruction functions
 */
kite_object_t *kite_new_null(kite_thread_t *);
kite_object_t *kite_new_integer(kite_thread_t *, long);
kite_object_t *kite_new_float(kite_thread_t *, double);
kite_object_t *kite_new_string(kite_thread_t *, char *);
kite_object_t *kite_new_string_with_length(kite_thread_t *, char *, int);
kite_object_t *kite_new_ident(kite_thread_t *, char *);
kite_object_t *kite_new_boolean(kite_thread_t *, int);
kite_object_t *kite_new_list(kite_thread_t *);
kite_object_t *kite_new_method(kite_thread_t *, int, void*, int);
kite_object_t *kite_new_method_with_docs(kite_thread_t *, int, void*, char*, int, ...);
kite_object_t *kite_new_class(kite_thread_t *, kite_object_t *, char *);
kite_object_t *kite_new_instance(kite_thread_t *, kite_object_t *);
kite_object_t *kite_new_instance_with_constructor(kite_thread_t *, 
    kite_object_t *, kite_object_t *);
kite_object_t *kite_new_exception(kite_thread_t *, char *, char *);
void kite_destruct_object(kite_thread_t *, kite_object_t **, int);
void kite_destruct_object_nofree(kite_thread_t *, kite_object_t **, int);
void kite_dereference_object(kite_object_t *);
kite_object_t *kite_reference_object(kite_object_t *);

/* 
 * Stuff to allow built-in and loadable extensions to add their own
 * properties and methods, among other things.
 */
void kite_add_property(kite_thread_t *, kite_object_t *, char *, int, char *);
void kite_add_method(kite_thread_t *, kite_object_t *, char *, kite_object_t *);
void kite_add_operator(kite_thread_t *, kite_object_t *, int, kite_object_t *);
void kite_set_docstring(kite_object_t *, char *);
char *kite_get_docstring(kite_object_t *);
void kite_set_arginfo(kite_object_t *, kite_object_t *);
void kite_remove_property(kite_thread_t *, kite_object_t *, char *, int);
void kite_remove_method(kite_thread_t *, kite_object_t *, char *, int);
void kite_remove_operator(kite_thread_t *, kite_object_t *, int);
void kite_set_property(kite_thread_t *, kite_object_t *, char *, kite_object_t *);
int kite_exists_property(kite_object_t *, char *);
int kite_exists_method(kite_object_t *, char *);
int kite_exists_operator(kite_object_t *, int);
kite_object_t *kite_get_property(kite_object_t *, char *);
kite_object_t *kite_dereference_and_load(kite_thread_t *, char *);
void kite_loader_register(kite_thread_t *, char *, void (*)(kite_thread_t*, kite_object_t*));

/*
 * Stuff to find properties/methods
 */
#define KITE_FIND_PROPERTY(ret, obj, name, num) \
    KITE_FIND_IN_SYMTAB(ret, obj->object_data.properties, name, num)
#define KITE_FIND_METHOD(ret, obj, name, num) \
    KITE_FIND_IN_SYMTAB(ret, obj->object_data.properties, name, num)
#define KITE_FIND_IN_SYMTAB(ret, symtab, n, num) \
    { \
        ret = symtab; \
        while(ret) { \
            int r = strcmp(n, ret->name); \
            if (!r && ret->global == num) { \
                break; \
            } else if (r < 0 || (!r && ret->global > num)) ret = ret->left; \
            else ret = ret->right; \
        } \
    }
#define KITE_FIND_ANY_IN_SYMTAB(ret, symtab, n) \
    { \
        ret = symtab; \
        while(ret) { \
            int r = strcmp(n, ret->name); \
            if (!r) { \
                break; \
            } else if (r < 0) ret = ret->left; \
            else ret = ret->right; \
        } \
    }
    
/*
 * Symbol table manipulation
 */ 
void kite_symtab_insert(kite_thread_t *, kite_symtab_t **, kite_symtab_t *);
void kite_symtab_remove(kite_thread_t *, kite_symtab_t **, char *, int);
void kite_copy_symtab(kite_symtab_t **, kite_symtab_t *);
void kite_destruct_symtab(kite_thread_t *, kite_symtab_t **, int);
void kite_copy_symtab_elements(kite_thread_t *, kite_symtab_t **, kite_symtab_t *, kite_object_t *);

/*
 * Canonical object representation
 */
kite_object_t *kite_object_name(kite_thread_t *, kite_object_t *);
kite_object_t *kite_boolean_object(kite_thread_t *, kite_object_t *);
kite_object_t *kite_string_object(kite_thread_t *, kite_object_t *);
kite_object_t *kite_int_object(kite_thread_t *, kite_object_t *);
kite_object_t *kite_float_object(kite_thread_t *, kite_object_t *);
kite_object_t *kite_list_object(kite_thread_t *, kite_object_t *);

/*
 * List manipulation
 */
int kite_list_count(kite_thread_t *, kite_object_t *);
void kite_append_list(kite_thread_t *, kite_object_t *, kite_object_t *);

/*
 * Defines for module writers
 */
#define KITE_MODULE_INITIALIZER_NAME(name) \
     name ## _load_module
#define KITE_MODULE_INITIALIZER(name) \
     void KITE_MODULE_INITIALIZER_NAME(name) (kite_thread_t *, kite_object_t *); \
     void* name ## _load_module__wrap(kite_thread_t *thd) { \
         kite_loader_register(thd, #name, KITE_MODULE_INITIALIZER_NAME(name)); \
         return NULL; \
     } \
     void KITE_MODULE_INITIALIZER_NAME(name) (kite_thread_t *thread, kite_object_t *parent)

#define KITE_CLASS_METHOD(name) \
     void name(kite_thread_t *thd, kite_object_t *this, \
               kite_object_t *args)
#define KITE_NO_ARGS (void)args;
#define KITE_GET_METHOD_ARGUMENT(ret, n) \
    { \
        int cur = n - 1; \
        ret = args; \
        while(ret && cur > 0) { \
            cur--; \
            ret = ret->builtin_data.listvalue.cdr; \
        } \
        if (ret) ret = ret->builtin_data.listvalue.car; \
    }
#define KITE_IS_TYPE(val, t) ((val)->type == t)

#define KITE_GET_INTEGER(v) ((v)->builtin_data.intvalue)
#define KITE_GET_FLOAT(v) ((v)->builtin_data.floatvalue)
#define KITE_GET_STRING_VALUE(v) ((v)->builtin_data.stringvalue.string)
#define KITE_GET_STRING_LENGTH(v) ((v)->builtin_data.stringvalue.length)
#define KITE_SET_INTEGER(v, i) KITE_GET_INTEGER(v) = i
#define KITE_SET_FLOAT(v, i) KITE_GET_FLOAT(v) = i

#endif /* KITE_OBJ__KITEOBJECT_H */
