/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#define KITE_EXECUTE_C
#include "objs/kite_object.h"
#include "kite_vm.h"
#include "kite_opcodes.h"

/*
 * Valid operator strings
 */
char *OPERATOR_STRINGS[] = {
    "+",
    "-",
    "*",
    "/",
    "%",
    "unary +",
    "unary -",
    "=",
    "==",
    "!=",
    "<",
    ">",
    "<=",
    ">=",
    "and",
    "or",
    "not",
    "xor",
    "<<",
    ">>",
    "map",
    "reduce",
    "[]",
    "|",
    ".",
    NULL,
};

char *OPERATOR_NAME_STRINGS[] = {
    "plus",
    "minus",
    "multiply",
    "divide",
    "mod",
    "unplus",
    "unminus",
    "",
    "equals",
    "nequals",
    "lt",
    "gt",
    "leq",
    "geq",
    "and",
    "or",
    "not",
    "xor",
    "lshift",
    "rshift",
    "map",
    "reduce",
    "array",
    "call",
    "property",
    "number_of_operators"
};

/*****************************************************************************
 * Kite opcode: Push onto execution stack
 ****************************************************************************/
static void 
kite_vm_execute_push(kite_thread_t *thd, kite_opcode_push *ins)
{
    assert(ins->obj != NULL);
    kite_vm_push(thd, ins->obj);
}

/*****************************************************************************
 * Kite opcode: Place function arguments into symbol table.
 ****************************************************************************/
static void 
kite_vm_execute_funcargs(kite_thread_t *thd, kite_opcode_funcargs *ins)
{
    kite_object_t *args, *cur_array, *cur_args, *arg;
    kite_symtab_t *symt;
    kite_funccall_info_t *stackentry;
    int deref_array = 0;
    assert(ins->args != NULL);
    
    args = kite_vm_pop(thd);
    if (args->type != OBJ_LIST) {
        deref_array = 1;
        args = kite_list_object(thd, args);
    }

    cur_array = args;
    cur_args = ins->args;
    
    while(cur_array && cur_args) {
        stackentry = (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
        arg = cur_args->builtin_data.listvalue.car;
        
        if (arg != NULL) {
            symt = malloc(sizeof(kite_symtab_t));
            symt->name = strdup(arg->builtin_data.stringvalue.string);
            symt->global = 0;
            symt->left = symt->right = NULL;
            symt->docstring = NULL;
            symt->value = 
                kite_reference_object(cur_array->builtin_data.listvalue.car);
            kite_symtab_insert(thd, &stackentry->locals, symt);
        }
    
        cur_array = cur_array->builtin_data.listvalue.cdr;
        cur_args = cur_args->builtin_data.listvalue.cdr;
    }

    if (deref_array) kite_dereference_object(args);
    assert(cur_array == NULL && cur_args == NULL);
}

/*****************************************************************************
 * Kite opcode: Dereference variable
 ****************************************************************************/
static void 
kite_vm_execute_deref_1(kite_thread_t *thd, kite_opcode_deref_1 *ins)
{
    kite_object_t *ident, *res = NULL, *exc;
    kite_stack_entry_t *cur = 
        &thd->func_stack->stack[thd->func_stack->length - 1];
    kite_symtab_t *found;
    kite_funccall_info_t *stackentry;
    int searchthis = 0;
    char *ident_string;
    assert(cur != NULL);
    
    ident = kite_vm_pop(thd);
    assert(ident->type == OBJ_IDENT);
    ident_string = ident->builtin_data.stringvalue.string;

    if (strcmp("this", ident_string) == 0) {
        res = kite_reference_object(((kite_funccall_info_t*)cur->obj)->this);
    } else if (strcmp("__root", ident_string) == 0) {
        res = kite_reference_object(thd->vm->root_package);
    } else if (strcmp("base", ident_string) == 0) {
        kite_symtab_t *cur_method;
        
        KITE_FIND_IN_SYMTAB(
            cur_method, ((kite_funccall_info_t*)cur->obj)->locals,
            "__current", 0
        );
        
        if (cur_method && cur_method->value && cur_method->value->parent)
        {
            res = kite_reference_object(cur_method->value->parent->object_data.inherit_from);
        }
        else
        {
            res = kite_new_null(thd);
        }
    } else {
        do {
            stackentry = (kite_funccall_info_t*)cur->obj;
            if (searchthis && stackentry->this) {
                KITE_FIND_ANY_IN_SYMTAB(found, 
                    stackentry->this->object_data.properties,
                    ident_string);
            } else {
                KITE_FIND_ANY_IN_SYMTAB(found, stackentry->locals,
                    ident_string);
            }
            if (found) {
                if (found->value) res = found->value;
                else res = found->value = kite_new_null(thd);
                break;
            } else {
                if (searchthis) {
                    searchthis = 0;
                    cur--;
                } else {
                    searchthis = 1;
                }
            }
        } while(cur >= thd->func_stack->stack);
        
        /* If not in the stack, find in the root namespace */
        if (res == NULL) {
            KITE_FIND_PROPERTY(found, thd->vm->root_package, 
                ident_string, 1);
            if (found) {
                res = found->value;
            }
        }
        
        if (res == NULL && ins->create) {
            /* Create entry in symbol table */
            stackentry = (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
            found = malloc(sizeof(kite_symtab_t));
            found->name = strdup(ident_string);
            found->global = 0;
            found->docstring = NULL;
            found->left = found->right = NULL;
            found->value = kite_new_null(thd);
            kite_symtab_insert(thd, &stackentry->locals, found);
            res = found->value;
        } else if (res == NULL && !ins->create) {
            char buf[1024];
            kite_funccall_info_t *fc = 
                kite_top_stack(thd->func_stack);
            PUSH_FUNC_STACK(NULL);
            entry->sys_or_user = fc->sys_or_user;
            
            sprintf(buf, "Could not find property or variable %s", 
                ident_string);
            exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                     buf);
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            POP_FUNC_STACK;
            return;
        }
        
        if (!found->value->object_data.inherit_from)
        {
            kite_object_t *sysobj =
                kite_dereference_and_load(thd, "System.object");
            if (found->value != sysobj)
                found->value->object_data.inherit_from = sysobj;
        }

        if (!found->value->shareable && found->value->owner_thread != thd) 
            goto thread_access_err;
        
        kite_reference_object(found->value);
        kite_push_stack(&thd->running_stack, TRUE, found);
        kite_dereference_object(ident);
        return;
    }

    /* Wrap it in a reference */
    if (!res->shareable && res->owner_thread != thd) 
        goto thread_access_err;
        
    kite_vm_push(thd, res);
    kite_dereference_object(res);
    kite_dereference_object(ident);
    return;

thread_access_err:
    {
        PUSH_FUNC_STACK(NULL);
        exc = 
            kite_new_exception(
                thd, 
                "System.exceptions.AccessDenied", 
                "Cannot access unshared object."
            );
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
}

/*****************************************************************************
 * Kite opcode: Define variable
 ****************************************************************************/
static void 
kite_vm_execute_defprop(kite_thread_t *thd, kite_opcode_defprop *ins)
{
    kite_object_t *ident;
    kite_symtab_t *found;
    kite_funccall_info_t *stackentry;

    (void)ins;
    ident = kite_vm_pop(thd);
    assert(ident->type == OBJ_IDENT);

    /* Create entry in symbol table */
    stackentry = (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
    KITE_FIND_IN_SYMTAB(found, stackentry->locals, ident->builtin_data.stringvalue.string, 0)
    if (found == NULL) {
        found = malloc(sizeof(kite_symtab_t));
        found->name = strdup(ident->builtin_data.stringvalue.string);
        found->global = 0;
        found->docstring = NULL;
        found->left = found->right = NULL;
        found->value = kite_new_null(thd);
        kite_symtab_insert(thd, &stackentry->locals, found);
    }

    kite_dereference_object(ident);
}

/*****************************************************************************
 * Kite opcode: Dereference object property
 ****************************************************************************/
static void 
kite_vm_execute_deref_2(kite_thread_t *thd, kite_opcode_deref_2 *ins)
{
    kite_object_t *ident, *obj, *res = NULL, *curr, *exc;
    kite_symtab_t *entry;
    char buf[1024];
    char *name;
    
    ident = kite_vm_pop(thd);
    obj = kite_reference_object(kite_vm_pop(thd));
    
    curr = obj;
    name = ident->builtin_data.stringvalue.string;
    
    do {
        KITE_FIND_ANY_IN_SYMTAB(entry, curr->object_data.properties, name);
        if (entry) {
            if ((!entry->global && obj != curr) ||
                (entry->value && entry->value->type == OBJ_METHOD && 
                 obj != curr)) {
                break;
            }
            
            if (!entry->value) entry->value = kite_new_null(thd);
            else kite_reference_object(entry->value);
            
            if (!entry->value->shareable && entry->value->owner_thread != thd) goto thread_access_err;
            
            kite_push_stack(&thd->running_stack, TRUE, entry);
            kite_dereference_object(ident);
            kite_dereference_object(obj);
            return;
        }
        
        curr = curr->object_data.inherit_from;
    } while(curr != NULL);
    
    if (curr) {
        kite_add_property(thd, obj, name, 0, NULL);
        
        if (entry->value && entry->value->type == OBJ_METHOD) {
            res = kite_new_method(thd, 0, NULL, 0);
            memcpy(&res->builtin_data.funcvalue,
                   &entry->value->builtin_data.funcvalue,
                   sizeof(kite_function_t));
            res->builtin_data.funcvalue.this = obj;
        } else {
            res = kite_new_null(thd);
        }
        kite_set_property(thd, obj, name, res);
        
        KITE_FIND_IN_SYMTAB(entry, obj->object_data.properties, name, 0);
        
        if (!entry->value->shareable && entry->value->owner_thread != thd) goto thread_access_err;
        
        kite_push_stack(&thd->running_stack, TRUE, entry);
        kite_dereference_object(res);
        kite_dereference_object(ident);
        kite_dereference_object(obj);
        return;
    }
    
    curr = kite_new_list(thd);
    curr->builtin_data.listvalue.car = kite_new_string(thd, name);
    curr->builtin_data.listvalue.cdr = NULL;
            
    if (kite_vm_call_operator(thd, obj, OP_PROPERTY, curr, 0)) {
        kite_dereference_object(curr);
        return;
    }
        
    kite_dereference_object(curr);
    
    {
        kite_funccall_info_t *fc = 
            kite_top_stack(thd->func_stack);
        PUSH_FUNC_STACK(obj);
        entry->sys_or_user = fc->sys_or_user;
        
        sprintf(buf, "Could not find property %s", name);
        exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                 buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

thread_access_err:
    {
        PUSH_FUNC_STACK(NULL);
        exc = 
            kite_new_exception(
                thd, 
                "System.exceptions.AccessDenied", 
                "Cannot access unshared object."
            );
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
}

/*****************************************************************************
 * Kite opcode: Call method
 ****************************************************************************/
static void 
kite_vm_execute_call(kite_thread_t *thd, kite_opcode_call *ins)
{
    kite_object_t *args, *method, *this, *exc, *tmp;
    int curstack;

    args = kite_vm_pop(thd);
    method = kite_reference_object(kite_vm_pop(thd));
    this = kite_vm_pop(thd);

    curstack = thd->running_stack->length - 1;

    if (method->type == OBJ_METHOD && method->builtin_data.funcvalue.this) {
        this = method->builtin_data.funcvalue.this;
    }
    
    kite_funccall_info_t *fc = 
        kite_top_stack(thd->func_stack);
    PUSH_FUNC_STACK(this);
    entry->sys_or_user = fc->sys_or_user;
    
    if (method->type == OBJ_METHOD) {
        if (kite_list_count(thd, args) != 
            method->builtin_data.funcvalue.numargs) {
            exc = kite_new_exception(thd, "System.exceptions.InvalidArgument", 
                                     "Specified arguments do not match.");
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            return;
        }
        kite_vm_call_object(thd, this, method, args);
    } else if (method->type == OBJ_IDENT) {
        kite_vm_call_method(thd, this, method->builtin_data.stringvalue.string,
                            args, TRUE);
    } else {
        assert(0);
    }
    
    kite_dereference_object(args);
    kite_dereference_object(method);

    if (thd->exception) return;

    /* Remove anything other than the return value from the top of the stack */
    if ((thd->running_stack->length - 1) != curstack) {
        kite_object_t *ret = kite_vm_pop(thd);
        while((thd->running_stack->length - 1) != curstack) {
            if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
                kite_vm_pop(thd);
            } else {
                tmp = kite_vm_pop(thd);
                kite_dereference_object(tmp);
            }
        }
        if (strcmp("__construct__", method->builtin_data.stringvalue.string) != 0)
            kite_vm_push(thd, ret);
        POP_FUNC_STACK;
    } else {
        POP_FUNC_STACK;
        if (strcmp("__construct__", method->builtin_data.stringvalue.string) != 0)
            kite_vm_return(thd, kite_new_null(thd));
    }
}

/*****************************************************************************
 * Kite opcode: Execute code
 ****************************************************************************/
static void 
kite_vm_execute_eval(kite_thread_t *thd, kite_opcode_eval *ins)
{
    kite_object_t *code;

    (void)ins;
    code = kite_vm_pop(thd);
    kite_vm_execute_user_method(thd, code->builtin_data.funcvalue.funcptr);
    kite_dereference_object(code);                
}

/*****************************************************************************
 * Kite opcode: Construct a list
 ****************************************************************************/
static void 
kite_vm_execute_list_cons(kite_thread_t *thd, kite_opcode_list_cons *ins)
{
    kite_object_t *car, *cdr, *res;

    (void)ins;

    cdr = kite_vm_pop(thd);
    car = kite_vm_pop(thd);
    res = kite_new_list(thd);
    
    if (cdr->type != OBJ_LIST) {
        res->builtin_data.listvalue.cdr = kite_list_object(thd, cdr);
        kite_dereference_object(cdr);
    } else {
        res->builtin_data.listvalue.cdr = cdr;
    }
    res->builtin_data.listvalue.car = kite_reference_object(car);
    
    kite_vm_push(thd, res);
    kite_dereference_object(res);
}

/*****************************************************************************
 * Kite opcode: Construct a list
 ****************************************************************************/
static void 
kite_vm_execute_list_cons_1(kite_thread_t *thd, kite_opcode_list_cons_1 *ins)
{
    kite_object_t *car, *res;
    
    (void)ins;

    car = kite_vm_pop(thd);
    res = kite_new_list(thd);
    res->builtin_data.listvalue.car = kite_reference_object(car);
    
    kite_vm_push(thd, res);
    kite_dereference_object(res);
}

/*****************************************************************************
 * Kite opcode: Construct a list
 ****************************************************************************/
static void 
kite_vm_execute_list_cons_0(kite_thread_t *thd, kite_opcode_list_cons_0 *ins)
{
    kite_object_t *res;
    
    (void)ins;

    res = kite_new_list(thd);
    
    kite_vm_push(thd, res);
    kite_dereference_object(res);
}

/*****************************************************************************
 * Kite opcode: Duplicate top of stack
 ****************************************************************************/
static void 
kite_vm_execute_dupe_top(kite_thread_t *thd, kite_opcode_dupe_top *ins)
{
    kite_object_t *top;
    kite_symtab_t *symt;

    if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
        symt = (kite_symtab_t*)(thd->running_stack->stack[thd->running_stack->length - 1].obj);
        kite_reference_object(symt->value);
        kite_push_stack(&thd->running_stack, TRUE, symt);
    } else {
        top = kite_vm_pop(thd);
        kite_vm_push(thd, top);
        kite_vm_push(thd, top);
        kite_dereference_object(top);
        kite_dereference_object(top);
    }
}
    
/*****************************************************************************
 * Kite opcode: Jump if true
 ****************************************************************************/
static void *
kite_vm_execute_jumptrue(kite_thread_t *thd, kite_opcode_jump_true *ins)
{
    kite_object_t *obj, *onstack;
    void *newpc;
    assert(ins->jumpto != NULL && thd->func_stack != NULL);

    onstack = kite_vm_pop(thd);
    obj = kite_boolean_object(thd, onstack);
    if (obj->builtin_data.intvalue)
    {
        newpc = ins->jumpto;
    }
    else
    {
        newpc = ins->common.next;
    }
    kite_dereference_object(obj);
    kite_dereference_object(onstack);
    return newpc;
}

/*****************************************************************************
 * Kite opcode: Jump if false
 ****************************************************************************/
static void *
kite_vm_execute_jumpfalse(kite_thread_t *thd, kite_opcode_jump_false *ins)
{
    kite_object_t *obj, *onstack;
    void *newpc;
    assert(ins->jumpto != NULL && thd->func_stack != NULL);
 
    onstack = kite_vm_pop(thd);
    obj = kite_boolean_object(thd, onstack);
    if (!obj->builtin_data.intvalue)
    {
        newpc = ins->jumpto;
    }
    else
    {
        newpc = ins->common.next;
    }
    kite_dereference_object(obj);
    kite_dereference_object(onstack);
    return newpc;
}

/*****************************************************************************
 * Kite opcode: Jump unconditional
 ****************************************************************************/
static void *
kite_vm_execute_jumpuncond(kite_thread_t *thd, kite_opcode_jump_unconditional *ins)
{
    assert(ins->jumpto != NULL && thd->func_stack != NULL);
    return ins->jumpto;
}

/*****************************************************************************
 * Kite opcode: Handle exception
 ****************************************************************************/
static void *kite_vm_execute_exception_hndl(
    kite_thread_t *thd, kite_opcode_exception_hndl *ins)
{
    int tos = thd->running_stack->length;
    int func_stack = thd->func_stack->length;
    
    kite_vm_execute_user_method(thd, ins->common.next);
    if (thd->exception)
    {
        while(thd->running_stack && thd->running_stack->length != tos) {
            kite_object_t *obj = kite_vm_pop(thd);
            kite_dereference_object(obj);
        }
        while(thd->func_stack && thd->func_stack->length != func_stack) {
            kite_funccall_info_t *obj = kite_pop_stack(&thd->func_stack);
            free(obj);
        }
        
        kite_symtab_t *propinfo = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
        assert(propinfo != NULL);
        propinfo->name = strdup("__exc");
        propinfo->global = 0;
        propinfo->docstring = NULL;
        propinfo->left = NULL;
        propinfo->right = NULL;
        propinfo->value = thd->exception;
        
        kite_funccall_info_t *current = 
            (kite_funccall_info_t*)(thd->func_stack->stack[thd->func_stack->length - 1].obj);
        kite_symtab_insert(thd, &current->locals, propinfo);
        
        thd->exception = NULL;
        return ins->handler;
    }
    
    return NULL;
}

/*****************************************************************************
 * Kite opcode: Handle exception
 ****************************************************************************/
static void kite_vm_execute_pop_exception_hndl(
    kite_thread_t *thd, kite_opcode_pop_exception_hndl *ins)
{
    kite_symtab_t *symt;
    (void)ins;
    
    kite_funccall_info_t *current = 
        (kite_funccall_info_t*)(thd->func_stack->stack[thd->func_stack->length - 1].obj);
        
    KITE_FIND_IN_SYMTAB(symt, current->locals, "__exc", 0);
    if (symt) kite_symtab_remove(thd, &current->locals, "__exc", 0);
}

/*****************************************************************************
 * Kite opcode: Get "this"
 ****************************************************************************/
static void kite_vm_execute_this(kite_thread_t *thd, kite_opcode_this *ins)
{
    (void)ins;
    kite_vm_push(thd, 
        ((kite_funccall_info_t*)kite_top_stack(thd->func_stack))->this);
}

/*****************************************************************************
 * Kite opcode: Execute arithmetic operation
 ****************************************************************************/
static void
kite_vm_execute_arithop(kite_thread_t *thd, kite_opcode_arithop *ins)
{
    kite_object_t *lhs, *rhs, *tmp ;

    assert(ins->operation < NUM_OPERATORS);
    rhs = kite_vm_pop(thd);

    if (thd->running_stack->stack[thd->running_stack->length - 1].reference &&
        ins->operation == OP_ASSIGNMENT) {
        kite_symtab_t *symt = 
            (kite_symtab_t*)thd->running_stack->stack[thd->running_stack->length - 1].obj;
        kite_object_t *oldobj = symt->value;
        symt->value = kite_reference_object(rhs);
        kite_dereference_object(oldobj);
        kite_dereference_object(oldobj->parent);
        if (oldobj->type != OBJ_CLASS)
            oldobj->parent = NULL;
        kite_vm_pop(thd);
        return;
    }
    lhs = kite_vm_pop(thd);

    /* 
     * Short-circuit, most common operations--we don't need to call the method.
     */ 
    if (lhs->type == rhs->type && (
            lhs->type == OBJ_INTEGER ||
            lhs->type == OBJ_FLOAT ||
            lhs->type == OBJ_BOOLEAN
        ) &&
        ins->operation != OP_ASSIGNMENT)
    {
        if (lhs->object_data.inherit_from->object_data.operators[ins->operation])
        {
            kite_compiled_func_t func =
                (lhs->object_data.inherit_from->object_data.operators[ins->operation]->builtin_data.funcvalue.funcptr);
            (*func)(thd, lhs, rhs);
            return;
        }
    }

    kite_funccall_info_t *fc = 
        kite_top_stack(thd->func_stack);
    PUSH_FUNC_STACK(lhs);
    entry->sys_or_user = fc->sys_or_user;

    int curstack = thd->running_stack->length - 1;
    kite_vm_call_operator(thd, lhs, ins->operation, rhs, TRUE);
    if (thd->exception) return;

    /* Remove anything other than the return value from the top of the stack */
    if ((thd->running_stack->length - 1) != curstack) {
        kite_object_t *ret = kite_vm_pop(thd);
        while((thd->running_stack->length - 1) != curstack) {
            if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
                kite_vm_pop(thd);
            } else {
                tmp = kite_vm_pop(thd);
                kite_dereference_object(tmp);
            }
        }
        kite_vm_push(thd, ret);
        POP_FUNC_STACK;
    } else {
        POP_FUNC_STACK;
        kite_vm_return(thd, kite_new_null(thd));
    }
}

/*****************************************************************************
 * Kite opcode: Execute make operation
 ****************************************************************************/
static void
kite_vm_execute_make(kite_thread_t *thd, kite_opcode_make *ins)
{
    kite_object_t *args, *obj, *newobj;
    int current_stack;

    args = kite_vm_pop(thd);
    obj = kite_vm_pop(thd);
    current_stack = thd->running_stack->length;

    kite_funccall_info_t *fc = 
        kite_top_stack(thd->func_stack);
    PUSH_FUNC_STACK(obj);
    entry->sys_or_user = fc->sys_or_user;
    
    newobj = kite_new_instance_with_constructor(thd, obj, args);
    if (thd->exception) return;
    POP_FUNC_STACK;

    if (thd->running_stack->length != current_stack) {
        while(thd->running_stack->length != current_stack) {
            obj = kite_vm_pop(thd);
            kite_dereference_object(obj);
        }
    }

    kite_vm_push(thd, newobj);
    kite_dereference_object(newobj);
}

/*****************************************************************************
 * Kite opcode: Execute is/isof operation
 ****************************************************************************/
static void
kite_vm_execute_objis(kite_thread_t *thd, kite_opcode_objis *ins)
{
    kite_object_t *class, *obj, *cur;
    int res = FALSE;
    
    class = kite_vm_pop(thd);
    obj = kite_vm_pop(thd);
    
    if (obj->type != OBJ_CLASS) {
        cur = obj->object_data.inherit_from;
    } else {
        cur = obj;
    }
    
    do {
        if (cur == class) res = TRUE;
        
        cur = cur->object_data.inherit_from;
    } while(cur && !res && ins->isof);

    obj = kite_new_boolean(thd, res);
    kite_vm_push(thd, obj);
    //kite_dereference_object(obj);
}

/*****************************************************************************
 * Call method on object programmatically
 ****************************************************************************/
void kite_vm_call_method(kite_thread_t *thd, kite_object_t *obj, 
                         char *name, struct kite_object_t *args, int err)
{
    kite_symtab_t *found = NULL;
    kite_object_t *curr = obj, *exc;
    char buf[1024];
    int numargs = kite_list_count(thd, args);
    
    do {
        KITE_FIND_METHOD(found, curr, name, numargs);
        if (found && found->value->type == OBJ_METHOD) {
            kite_vm_call_object(thd, obj, found->value, args);
            return;
        } else if (found) {
            sprintf(buf, "%s is not callable.", name);
            exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                                     buf);
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            return;
        }
        
        curr = curr->object_data.inherit_from;
    } while(curr);
    
    /* Call function operator if method's not found */
    if (strcmp("__construct__", name) != 0) {
        curr = kite_new_list(thd);
        curr->builtin_data.listvalue.car = kite_new_string(thd, name);
        curr->builtin_data.listvalue.cdr = kite_new_list(thd);
        curr->builtin_data.listvalue.cdr->builtin_data.listvalue.car =
            kite_reference_object(args);
            
        if (kite_vm_call_operator(thd, obj, OP_METHOD_CALL, curr, 0)) {
            kite_dereference_object(curr);
            return;
        }
        
        kite_dereference_object(curr);
    }
    
    /* Method not found, raise exception */
    if (err) {
        if (strcmp("__construct__", name) == 0) {
            sprintf(buf, "Could not find constructor that takes %d argument(s)",
                    numargs);
        } else {
            sprintf(buf, "Could not find method %s that takes %d argument(s)",
                    name, numargs);
        }
        exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                 buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
}

/*****************************************************************************
 * Call method on object programmatically
 ****************************************************************************/
void 
kite_vm_call_object(kite_thread_t *thd, struct kite_object_t *this, 
                    struct kite_object_t *method, struct kite_object_t *args)
{
    kite_compiled_func_t func;
    int freestack = 0;
    kite_object_t *oldthis = NULL;
    
    /*
     * [Ticket 5] Replace with current this if we're simply calling a parent 
     * class's method.
     */
    if (this->type != OBJ_INSTANCE && thd->func_stack && 
        thd->func_stack->length > 1) {
        kite_funccall_info_t *current = 
            (kite_funccall_info_t*)(thd->func_stack->stack[thd->func_stack->length - 2].obj);
        kite_object_t *cur_this = current->this;
        while(cur_this != NULL) {
            if (cur_this == this && current->this->type == OBJ_INSTANCE) {
                oldthis = this;
                this = current->this;
                break;
            }
            cur_this = cur_this->object_data.inherit_from;
        }
    }
    
    assert(method->type == OBJ_METHOD || method->type == OBJ_CLASS ||
           method->type == OBJ_INSTANCE);
    if (method->builtin_data.funcvalue.functype == FUNC_COMPILED) {
        func = method->builtin_data.funcvalue.funcptr;
        func(thd, this, args);
        return;
    }
    
    /* Non-compiled method */
    if (!thd->func_stack || thd->func_stack->length == 0) {
        freestack = 1;
        kite_funccall_info_t *entry = calloc(sizeof(kite_funccall_info_t), 1);
        entry->this = this;
        entry->file = "(Kite loader)";
        kite_push_stack(&thd->func_stack, FALSE, entry);
    } else {
        kite_funccall_info_t *entry = 
            (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
        entry->this = this;
    }
    
    // Request by ozzloy @ irc.freenode.net/#sdcolleges 6/5/2008
    // __caller variable will point to currently running method.
    // Perfect for anonymous recursive functions.
    kite_symtab_t *propinfo;
    
    propinfo = (kite_symtab_t*)malloc(sizeof(kite_symtab_t));
    assert(propinfo != NULL);
    propinfo->name = strdup("__current");
    propinfo->global = 0;
    propinfo->docstring = NULL;
    propinfo->left = NULL;
    propinfo->right = NULL;
    propinfo->value = kite_reference_object(method);
    
    kite_funccall_info_t *current = 
        (kite_funccall_info_t*)(thd->func_stack->stack[thd->func_stack->length - 1].obj);
    kite_symtab_insert(thd, &current->locals, propinfo);
    
    // Push arguments onto stack and begin executing function.
    kite_vm_push(thd, args);
    kite_vm_execute_user_method(thd, method->builtin_data.funcvalue.funcptr);
    
    if (oldthis) {
        kite_funccall_info_t *entry = 
            (kite_funccall_info_t*)kite_top_stack(thd->func_stack);
        entry->this = oldthis;
    }

    if (freestack) {
        kite_funccall_info_t *f = kite_top_stack(thd->func_stack);
        if (f && f->locals) kite_destruct_symtab(thd, &f->locals, 0);
        free(kite_pop_stack(&thd->func_stack));
    }
}

/*****************************************************************************
 * Call constructor on object programmatically
 ****************************************************************************/
void kite_vm_call_constructor(kite_thread_t *thd, struct kite_object_t *this, 
                              struct kite_object_t *args) 
{
    kite_symtab_t *symt;
    kite_object_t *curr = this->object_data.inherit_from;
    
    /* 
     * Find __construct__ (with any arguments) in the object first. If we
     * can't find any, just ignore this call. Otherwise, call constructor
     * and have the system throw an exception if the arguments don't match
     * up.
     *
     * Note [Ticket 5]: we're only looking in the current object, not the
     * parents. Any constructors in the current object are responsible for
     * calling parent constructors as needed (by using 
     * classname.__construct__).
     *
     * Note2: We also want to spit out an error if we're trying to call a
     * constructor that takes arguments that does not exist.
     */    
    KITE_FIND_ANY_IN_SYMTAB(symt, curr->object_data.properties, 
        "__construct__");
    if (symt != NULL || kite_list_count(thd, args) > 0) {
        kite_vm_call_method(thd, this, "__construct__", args, TRUE);
        return;
    }
}

/*****************************************************************************
 * Call operator method on object programmatically
 ****************************************************************************/
int kite_vm_call_operator(kite_thread_t *thd, struct kite_object_t *this, 
                          int op, struct kite_object_t *arg, int err)
{
    kite_object_t *cur = this, *exc;
    
    do {
        if (cur->object_data.operators[op]) {
            kite_vm_call_object(thd, this, cur->object_data.operators[op], 
                                arg);
            return 1;
        }
        
        cur = cur->object_data.inherit_from;
    } while(cur);
    
    /* Operator not defined, raise exception */
    if (err) {
        exc = kite_new_exception(thd, "System.exceptions.OperatorNotDefined", 
                                 OP_TO_STRING(op));
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    return 0;
}

/*****************************************************************************
 * Handle exception thrown in user code or Kite standard library/VM
 ****************************************************************************/
void kite_vm_execute_exception(kite_thread_t *thd)
{
    kite_object_t *name = kite_object_name(thd, thd->exception);
    kite_symtab_t *msg, *trace;
    kite_object_t *cur = NULL;
    
    KITE_FIND_PROPERTY(msg, thd->exception, "msg", 0);
    KITE_FIND_PROPERTY(trace, thd->exception, "trace", 0);
    
    printf("%s: %s\n", name->builtin_data.stringvalue.string,
                       msg->value->builtin_data.stringvalue.string);
    if (trace) cur = trace->value;
    while(cur != NULL) {
        printf("  %s\n", 
            cur->builtin_data.listvalue.car->builtin_data.stringvalue.string);
        cur = cur->builtin_data.listvalue.cdr;
    }               
    kite_dereference_object(name);

    pthread_exit(0);
}

/*****************************************************************************
 * Main execution loop for Kite opcodes
 ****************************************************************************/
void *kite_vm_execute(void *ptr)
{
    kite_thread_t *thd = (kite_thread_t*)ptr;
    kite_object_t *args;
    
    if (thd->running_stack && thd->running_stack->length) {
        args = kite_vm_pop(thd);
    } else {
        args = kite_new_null(thd);
    }
    
    kite_vm_call_object(thd, thd->start, thd->start, args);
    kite_dereference_object(args);
    
    if (thd->exception) kite_vm_execute_exception(thd);

    return NULL;
}

/*****************************************************************************
 * Main execution loop for Kite opcodes
 ****************************************************************************/
void kite_vm_execute_user_method(kite_thread_t *thd, void *pc)
{
    kite_opcode_t *cur = (kite_opcode_t*)pc;
    while(cur)
    {
        kite_opcode_call *ins = (kite_opcode_call*)cur;
        kite_funccall_info_t *fc = 
            kite_top_stack(thd->func_stack);
            
        if (!fc || !fc->sys_or_user)
        {
            kite_symtab_t *st;
            KITE_FIND_PROPERTY(st, thd->vm->root_package, "__instruction_tracer", 1);
            if (st && st->value && st->value->type == OBJ_METHOD)
            {
                int curstack = thd->running_stack->length - 1;
                
                PUSH_FUNC_STACK(fc->this);
                entry->sys_or_user = 1;
                
                /* Function arguments */
                kite_object_t *args = kite_new_list(thd);
                kite_append_list(thd, args, kite_new_integer(thd, cur->opcode));
                kite_append_list(thd, args, kite_new_string(thd, cur->file));
                kite_append_list(thd, args, kite_new_integer(thd, cur->line));
                kite_object_t *usethis = st->value->builtin_data.funcvalue.this;
                kite_vm_call_object(thd, usethis ? usethis : fc->this, st->value, args);
                kite_dereference_object(args);
                
                /* Ignore return value and anything else pushed onto the stack. */
                if ((thd->running_stack->length - 1) != curstack) {
                    kite_object_t *ret = kite_vm_pop(thd);
                    while((thd->running_stack->length - 1) != curstack) {
                        if (thd->running_stack->stack[thd->running_stack->length - 1].reference) {
                            kite_vm_pop(thd);
                        } else {
                            kite_object_t *tmp = kite_vm_pop(thd);
                            kite_dereference_object(tmp);
                        }
                    }
                    kite_dereference_object(ret);
                }
                
                POP_FUNC_STACK;
            }
        }
        
        /* Decode, execute */
        switch(cur->opcode)
        {
            case ARITH_OP:
            {
                kite_vm_execute_arithop(thd, (kite_opcode_arithop*)cur);
                cur = cur->next;
                break;
            }
            case PUSH:
            {
                kite_vm_execute_push(thd, (kite_opcode_push*)cur);
                cur = cur->next;
                break;
            }
            case LIST_CONS:
            {
                kite_vm_execute_list_cons(thd, (kite_opcode_list_cons*)cur);
                cur = cur->next;
                break;
            }
            case LIST_CONS_0:
            {
                kite_vm_execute_list_cons_0(thd, (kite_opcode_list_cons_0*)cur);
                cur = cur->next;
                break;
            }
            case LIST_CONS_1:
            {
                kite_vm_execute_list_cons_1(thd, (kite_opcode_list_cons_1*)cur);
                cur = cur->next;
                break;
            }
            case DEREF_1:
            {
                kite_vm_execute_deref_1(thd, (kite_opcode_deref_1*)cur);
                cur = cur->next;
                break;
            }
            case DEREF_2:
            {
                kite_vm_execute_deref_2(thd, (kite_opcode_deref_2*)cur);
                cur = cur->next;
                break;
            }
            case CALL:
            {
                kite_vm_execute_call(thd, (kite_opcode_call*)cur);
                cur = cur->next;
                break;
            }
            case JUMP_TRUE:
            {
                cur = 
                    kite_vm_execute_jumptrue(thd, (kite_opcode_jump_true*)cur);
                break;
            }
            case JUMP_FALSE:
            {
                cur = 
                    kite_vm_execute_jumpfalse(thd, (kite_opcode_jump_false*)cur);
                break;
            }
            case JUMP_UNCOND:
            {
                cur = 
                    kite_vm_execute_jumpuncond(thd, (kite_opcode_jump_unconditional*)cur);
                break;
            }
            case EXCEPTION_HNDL:
            {
                cur = kite_vm_execute_exception_hndl(
                    thd, (kite_opcode_exception_hndl*)cur);
                break;
            }
            case POP_EXCEPTION_HNDL:
            {
                kite_vm_execute_pop_exception_hndl(
                    thd, (kite_opcode_pop_exception_hndl*)cur);
                cur = cur->next;
                break;
            }
            case DUPE_TOP:
            {
                kite_vm_execute_dupe_top(thd, (kite_opcode_dupe_top*)cur);
                cur = cur->next;
                break;
            }
            case FUNCARGS:
            {
                kite_vm_execute_funcargs(thd, (kite_opcode_funcargs*)cur);
                cur = cur->next;
                break;
            }
            case MAKE_OBJ:
            {
                kite_vm_execute_make(thd, (kite_opcode_make*)cur);
                cur = cur->next;
                break;
            }
            case EVAL_OBJ:
            {
                kite_vm_execute_eval(thd, (kite_opcode_eval*)cur);
                cur = cur->next;
                break;
            }
            case IS_ISOF:
            {
                kite_vm_execute_objis(thd, (kite_opcode_objis*)cur);
                cur = cur->next;
                break;
            }
            case DEFPROP:
            {
                kite_vm_execute_defprop(thd, (kite_opcode_defprop*)cur);
                cur = cur->next;
                break;
            }
            case PUSH_THIS:
            {
                kite_vm_execute_this(thd, (kite_opcode_this*)cur);
                cur = cur->next;
                break;
            }
            case NOP:
            {
                /* Ignore. */
                cur = cur->next;
                break;
            }
            case RETURN_NOW:
            {
                /* Exit out of method. */
                kite_dereference_object(thd->exception);
                thd->exception = NULL;
                return;
            }
            default:
            {
                /* Invalid opcode */
                assert(0);
                break;
            }
        }

        if (thd->exception) return;
    }
}
